package modem

import (
	"encoding/json"
	"errors"
	"fmt"
	"log"
	"math"
	"os"
	"os/exec"
	"slices"
	"strconv"
	"strings"
	"time"

	"github.com/CGSG-2021-AE4/modem-test/api/modem/at"
)

type GpsInfo struct {
	Latitude           float64 `json:"Latitude"`
	Longitude          float64 `json:"Longitude"`
	LatitudeIndicator  string  `json:"Latitude_indicator"`  // North/South
	LongitudeIndicator string  `json:"Longitude_indicator"` // West/East
	Speed              float64 `json:"Speed"`
	Course             float64 `json:"-"`
	Altitude           float64 `json:"-"`
	Date               string  `json:"-"`
	Time               string  `json:"-"`
}

var GpsInfoNil = GpsInfo{}

type modem struct {
	// Internal values
	logger *log.Logger

	// Serial values
	baudrate int

	deviceName  string
	port        at.Port
	isAvailable bool

	// Gpio values
	onOffPin gpioPin

	// Other values
	gpsInfo        GpsInfo
	lastUpdateTime time.Time
}

type Modem interface {
	Init() error
	SearchPort(isSoft bool) error
	Connect() error
	Ping() error
	SwitchToGpsMode() error
	CalculateSpeed(newLatitude, newlongitude float64)
	Update() error
	GetInfo() (string, error)
	TestGPS() error
}

func New() Modem {
	return &modem{
		logger:         log.New(os.Stdout, "modem:", log.LstdFlags),
		baudrate:       115200,
		onOffPin:       gpioPin{Pin: 6},
		lastUpdateTime: time.Now(),
	}
}

func (m *modem) Init() error {
	// Turn module on
	// log.Println("=============================== Turn on module")
	// if err := m.onOffPin.Init(); err != nil {
	// 	return fmt.Errorf("gpio pin init: %w", err)
	// }
	// onOffPin.PowerOn()

	// Search
	m.logger.Println("=============================== Search")
	// Soft search
	if err := m.SearchPort(true); err != nil {
		return fmt.Errorf("soft port search: %w", err)
	}
	// Wide search
	if m.port == nil {
		if err := m.SearchPort(false); err != nil {
			return fmt.Errorf("not soft port search: %w", err)
		}
	}
	if m.port == nil {
		return errors.New("no port is detected")
	}

	// Connect
	m.logger.Println("=============================== Connect")
	if err := m.Connect(); err != nil {
		return fmt.Errorf("connect: %w", err)
	}

	// Tests
	m.logger.Println("=============================== Test")
	if err := m.TestGPS(); err != nil {
		return fmt.Errorf("testGPS: %w", err)
	}
	return nil
}

func (m *modem) checkPort(portName string) error {
	// Check device for existance
	if _, err := os.Stat(portName); err != nil {
		return fmt.Errorf("device does not exist")
	}

	// Check serial connection
	// Connect
	m.port = at.New(portName, m.baudrate)
	if err := m.port.Connect(); err != nil {
		return fmt.Errorf("connect: %w", err)
	}
	defer m.port.Disconnect() // Do not bother about errors...

	// Ping
	m.logger.Println("Ping...")

	if err := m.Ping(); err != nil {
		return fmt.Errorf("ping error: %w", err)
	}

	// Check model
	m.logger.Println("Check model...")

	resp, err := m.port.Send("AT+CGMM")
	if err != nil {
		return fmt.Errorf("get model: %w", err)
	}
	if !resp.Check() {
		return fmt.Errorf("error response: %s", resp)
	}
	model := strings.Split(resp.String(), "\n")[0]
	if err != nil {
		return fmt.Errorf("get model: %w", err)
	}
	rightModel := "SIMCOM_SIM7600E-H"
	// m.logger.Printf("[% x]\n [% x]", []byte("SIMCOM_SIM7600E-H"), []byte(model))
	if model[:len(rightModel)] != rightModel {
		return fmt.Errorf("invalid modem model: %s", model)
	}
	m.logger.Println("Model right")
	return nil
}

func (m *modem) SearchPort(isSoft bool) error {
	// Get ports
	ports := []string{"ttyUSB1", "ttyUSB2", "ttyUSB3"}
	if !isSoft {
		ps, err := GetTtyDevices()
		if err != nil {
			fmt.Errorf("get serial devices: %w", err)
		}
		ports = ps
	}

	// Check ports
SearchLoop:
	for _, p := range ports {
		m.logger.Printf("Checking port %s ...\n", p)

		if err := m.checkPort("/dev/" + p); err != nil {
			m.logger.Printf("Check failed: %s\n", err.Error())
			continue SearchLoop
		}

		m.logger.Print("Found modem on port: ", p)
		m.port = at.New("/dev/"+p, m.baudrate)
		m.isAvailable = true
		return nil
	}
	return nil
}

func (m *modem) Connect() error {
	if m.port == nil {
		return fmt.Errorf("port is not defined")
	}
	return m.port.Connect()
}

func (m *modem) Ping() error {
	resp, err := m.port.Send("AT")
	if err != nil {
		return err
	}
	if !resp.Check() {
		return fmt.Errorf("connection lost")
	}
	return nil
}

func (m *modem) SwitchToGpsMode() error {
	m.logger.Println("Enabling GPS mode...")
	// Reset intput
	if err := m.port.GetSerialPort().ResetInputBuffer(); err != nil {
		return fmt.Errorf("reset input buffer: %w", err)
	}

	// Check gps mode status
	resp, err := m.port.Send("AT+CGPS?")
	if err != nil {
		return fmt.Errorf("make at ask: %w", err)
	}
	if !resp.Check() {
		return fmt.Errorf("error response")
	}
	ans := strings.Replace(strings.Split(strings.Split(resp.RmFront("+CGPS:").String(), "\n")[0], ",")[0], " ", "", -1)
	if ans == "1" {
		m.logger.Println("GPS already enabled")
		return nil
	}
	m.logger.Println(ans)

	// Modem is not in GPS mode
	resp, err = m.port.Send("AT+CGPS=1")
	if err != nil {
		return fmt.Errorf("try to switch to gps: %w", err)
	}
	if !resp.Check() {
		return fmt.Errorf("switch tp GPS failed")
	}
	m.logger.Println("GPS mode enabled")
	return nil
}

func deg2rad(deg float64) float64 {
	return deg * (math.Pi / 180)
}

func (m *modem) CalculateSpeed(newLatitude, newLongitude float64) {
	m.logger.Println("Calculate speed")
	earthRad := 6371.0 // TODO ?
	dLat := deg2rad(math.Abs(newLatitude - m.gpsInfo.Latitude))
	dLon := deg2rad(math.Abs(newLongitude - m.gpsInfo.Longitude))
	a := math.Sin(dLat/2)*math.Sin(dLat/2) +
		math.Cos(deg2rad(newLatitude))*math.Cos(deg2rad(m.gpsInfo.Latitude))*math.Sin(dLon/2)*math.Sin(dLon/2)
	c := 2 * math.Atan2(math.Sqrt(a), math.Sqrt(1-a))

	m.gpsInfo.Speed = earthRad * c / (math.Abs(float64(time.Since(m.lastUpdateTime))))
}

func decodeGpsInfo(str string) (GpsInfo, error) {
	var err error
	newGpsInfo := GpsInfo{}
	strs := strings.Split(strings.Split(strings.Replace(str, " ", "", -1), "\n")[0], ",")

	newGpsInfo.Latitude, err = strconv.ParseFloat(strs[0], 64)
	if err != nil {
		return GpsInfoNil, fmt.Errorf("parse latitude: %w", err)
	}
	newGpsInfo.Longitude, err = strconv.ParseFloat(strs[2], 64)
	if err != nil {
		return GpsInfoNil, fmt.Errorf("parse longitude: %w", err)
	}
	newGpsInfo.LatitudeIndicator = strs[1]
	newGpsInfo.LatitudeIndicator = strs[3]
	newGpsInfo.Date = strs[4]
	newGpsInfo.Time = strs[5]
	newGpsInfo.Altitude, err = strconv.ParseFloat(strs[6], 64)
	if err != nil {
		return GpsInfoNil, fmt.Errorf("parse altitude: %w", err)
	}
	newGpsInfo.Speed, err = strconv.ParseFloat(strs[7], 64)
	if err != nil {
		return GpsInfoNil, fmt.Errorf("parse speed: %w", err)
	}
	// Course sometimes may be null
	if len(strs[8]) > 0 {
		newGpsInfo.Course, err = strconv.ParseFloat(strs[8], 64)
		if err != nil {
			return GpsInfoNil, fmt.Errorf("parse course: %w", err)
		}
	}
	return newGpsInfo, nil
}

func (m *modem) Update() error {
	m.logger.Println("Update")
	if !m.isAvailable {
		m.logger.Println("No connection to module")
		return nil
	}
	// ans, err := m.port.Request(at.CmdQuestion, "CGPSINFO")
	// if err != nil {
	// 	return fmt.Errorf("check GPS info mode: %w", err)
	// }
	// ok := ans == "1"
	// if !ok {
	// 	_, err := m.port.Request(at.CmdCheck, "CGPSINFO")
	// 	if err != nil {
	// 		return fmt.Errorf("switch to GPS info mode: %w", err)
	// 	}
	// 	m.logger.Println("switched to GPS mode")
	// } else {
	// 	m.logger.Println("mode in right GPS mode")
	// }

	// Update
	m.logger.Println("Receiving GPS data...")
	resp, err := m.port.Send("AT+CGPSINFO")
	if err != nil {
		return fmt.Errorf("receive GPS data: %w", err)
	}
	if !resp.Check() {
		return fmt.Errorf("error response")
	}
	m.logger.Println("Decoding data...")

	newGpsInfo, err := decodeGpsInfo(strings.Split(strings.Replace(resp.RmFront("+CGPSINFO:").String(), "\r", "", -1), "\n")[0])
	if err != nil {
		m.logger.Println("Gps info decode error:", err.Error())
		return nil
	}
	m.gpsInfo = newGpsInfo
	m.logger.Println("Decoded successfully")
	return nil
}

type ModemInfo struct {
	Port string `json:"Port"`
	GpsInfo
}

type Info struct {
	Modem ModemInfo `json:"Modem"`
}

func (m *modem) GetInfo() (string, error) {
	info := Info{
		Modem: ModemInfo{
			Port:    m.port.GetName(),
			GpsInfo: m.gpsInfo,
		},
	}
	buf, err := json.Marshal(info)
	if err != nil {
		fmt.Errorf("marshal info: %w", err)
	}
	return string(buf), nil // why you clamp
}

// Difference: I do not set \n at the end of string
func (m *modem) GetShortInfo() string {
	return fmt.Sprintf("%f,%s,%f,%s", m.gpsInfo.Latitude, m.gpsInfo.LatitudeIndicator, m.gpsInfo.Longitude, m.gpsInfo.LongitudeIndicator)
}

func (m *modem) SaveGPS(path string) error {
	f, err := os.OpenFile(path, os.O_APPEND|os.O_WRONLY|os.O_CREATE, 0600)
	if err != nil {
		return fmt.Errorf("open file: %w", err)
	}

	defer f.Close()

	if _, err = f.WriteString(m.GetShortInfo()); err != nil {
		return fmt.Errorf("write file: %W", err)
	}
	return nil
}

func (m *modem) TestGPS() error {
	m.logger.Println("Testing GPS")

	if err := m.SwitchToGpsMode(); err != nil {
		return fmt.Errorf("switch to GPS: %w", err)
	}

	if err := m.Update(); err != nil {
		return fmt.Errorf("update: %w", err)
	}

	m.logger.Println("Current coords:", m.GetShortInfo())
	return nil
}

func GetTtyDevices() ([]string, error) {
	devices := []string{}

	// Get ports
	/**/
	out, err := exec.Command("/bin/ls", "/dev").Output()
	if err != nil {
		return nil, fmt.Errorf("execute ls command: %w", err)
	}
	allPorts := strings.Split(string(out), "\n")
	for _, p := range allPorts {
		if len(p) > 3 && p[:3] == "tty" {
			devices = append(devices, p)
		}
	}
	slices.Reverse(devices) // ASK why

	return devices, nil
}

/*
	TODOs:
		maybe to store read/write buf in obj
	QUESTIONS:
		JSON why you clamp
*/

/*
 */
